//
//  AppController.m
//  Regex Visualizer
//
//  Created on 7/24/09.
//  Copyright 2009 Michael Sanders. All rights reserved.
//  All code released under the MIT license.
//

#import "AppController.h"
#import "TransitionTableData.h"
#import "MyTableView.h"
#import "NSTextFieldFormat.h"
#import "PyToObjc.h"

@interface AppController (Private)

- (PyObject *)getPythonFunction:(const char *)function_name
               fromModuleObject:(PyObject *)module
                       withName:(NSString *)module_name;

@end

@implementation AppController

- (void)awakeFromNib
{
    NSBundle *mainBundle = [NSBundle mainBundle];
    NSString *modulePath = [mainBundle pathForResource:@"regex_parser"
                                                ofType:@"py"];

    // Display a helpful error message if the Python module is not readable.
    if (modulePath == nil) {
        NSString *message = [NSString stringWithFormat:
                             @"Regex Visualizer cannot work without "
                             @"regex_parser.py. It SHOULD be located at “%@”, "
                             @"but it doesn’t seem to be. Sorry!", modulePath];

        NSRunAlertPanel(@"Couldn’t run regex_parser.py", message,
                        @"Quit", nil, nil);
        [NSApp terminate:nil];
    }

    Py_Initialize();

    // For some reason to import Python modules by an absolute path, we have to
    // add their directory to Python's search paths, and then import the module
    // as we would a system module (without the extension).
    PyObject *sys_path = PySys_GetObject("path");
    PyList_Insert(sys_path, 0, Py_None);
    const char *resourcePath = [[mainBundle resourcePath] UTF8String];
    PyList_SetItem(sys_path, 0, PyString_FromString(resourcePath));

    PyObject *module = PyImport_Import(PyString_FromString("regex_parser"));
    NSAssert(module != NULL, @"Could not import module");

    PyBuildRegexFunction = [self getPythonFunction:"build_regex"
                                  fromModuleObject:module
                                          withName:@"regex_parser"];
    PyParseWordFunction = [self getPythonFunction:"parse_word"
                                 fromModuleObject:module
                                         withName:@"regex_parser"];
    Py_DECREF(module);

    [self changeRegex:nil];
    [stringField setDelegate:self];
    [explanatoryText setStringValue:@"Enter a string to parse."];
}

- (void)dealloc
{
    Py_XDECREF(PyBuildRegexFunction);
    Py_XDECREF(PyParseWordFunction);
    Py_XDECREF(PyTransitions);
    Py_Finalize();
    [super dealloc];
}

- (IBAction)displayHelp:(id)sender
{
    NSRunAlertPanel(@"Help!",
                    @"Regex Visualizer currently only supports formal regex "
                    @"notation; that is, the +, *, (,  and ) operators.",
                    @"Okay, thanks.", nil, nil);
}

#pragma mark Python stuff

// Attempts to return Python function in module and exits program with error
// message if unable to.
// Returns a new reference.
- (PyObject *)getPythonFunction:(const char *)function_name
               fromModuleObject:(PyObject *)module
                       withName:(NSString *)module_name
{
    if (PyObject_HasAttrString(module, function_name)) {
        PyObject *function = PyObject_GetAttrString(module, function_name);
        if (function == NULL || !PyCallable_Check(function)) {
            NSString *title = [NSString stringWithFormat:@"Couldn’t run %@.py",
                               module_name];
            NSRunAlertPanel(title,
                            @"Couldn’t call %s function in %@.py",
                            @"Quit", nil, nil, function_name, module_name);
            Py_XDECREF(function);
            exit(1);
        }
        return function;
    } else {
        NSString *title = [NSString stringWithFormat:@"Couldn’t run %@.py",
                           module_name];
        NSRunAlertPanel(title,
                        @"Couldn’t find %s function in %@.py",
                        @"Quit", nil, nil, function_name, module_name);
        exit(1);
    }
}

#pragma mark -

- (IBAction)changeRegex:(NSTextField *)sender
{
    const char *infixRegex = [[regexField stringValue] UTF8String];
    PyObject *PyString = PyString_FromString(infixRegex);
    // if (PyString == NULL) return;

    PyTransitions = PyObject_CallFunction(PyBuildRegexFunction, "O", PyString);
    Py_DECREF(PyString);
    NSArray *transitions = PyObjectToObjc(PyTransitions);

    if (transitions == nil || [transitions count] == 0) {
        invalidExpression = YES;
        NSBeep();
        [explanatoryText setStringValue:@"Invalid expression."];
        [transitionTableData setTransitions:nil];
        [tableView reloadData];
    } else {
        invalidExpression = NO;
        [transitionTableData setTransitions:transitions];
        [tableView reloadData];
        if (sender != nil) {
            [self parseWord];
        }
    }
}

#pragma mark stringField delegate

- (void)controlTextDidChange:(NSNotification *)note
{
    [self parseWord];
}

#pragma mark -

- (void)parseWord
{
    if (invalidExpression) return;

    NSString *string = [stringField stringValue];
    PyObject *PyString = PyString_FromString([string UTF8String]);
    if (PyString == NULL) return;

    PyObject *tuple = PyObject_CallFunction(PyParseWordFunction, "OO",
                                            PyTransitions, PyString);
    if (tuple == NULL) return;
    Py_DECREF(PyString);

    NSArray *retList = PyObjectToObjc(tuple);
    Py_DECREF(tuple);
    NSAssert(retList != nil, @"Could not create retList");

    int retVal = [[retList objectAtIndex:0] intValue];
    NSArray *states = [retList objectAtIndex:1];

    if ([string length] == 0) string = @"ε";
    if (retVal == 0) {
        [explanatoryText setStringWithFormat:@"%@ was accepted!", string];
    } else if (retVal == -1) {
        [explanatoryText setStringWithFormat:@"%@ was NOT accepted.", string];
    } else if (retVal == -2) {
        NSString *statesString = [states componentsJoinedByString:@", "];
        [explanatoryText setStringWithFormat:@"%@ died on state(s) %@",
                                              string, statesString];

        [tableView highlightCellsAtColumnIdentifier:[retList objectAtIndex:2]
                                               rows:states
                                         successRow:-1];
        return;
    }

    [tableView highlightCellsAtColumnIndex:0
                                      rows:states
                                successRow:[tableView numberOfRows] - 1];
}

@end
