//
//  PyToObjc.m
//  Regex Visualizer
//
//  Created on 7/27/09.
//  Copyright 2009 Michael Sanders. All rights reserved.
//  All code released under the MIT license.
//

#import "PyToObjc.h"

// Converts a PyObject to its corresponding Objective-C object;
// currently supports lists, tuples, sets, dictionaries, strings, booleans,
// integers, longs, and doubles.
id PyObjectToObjc(PyObject *object)
{
    if (object == NULL) {
        NSLog(@"WTF? Object is NULL!");
        return nil;
    } else if (PyString_Check(object)) {
        return [NSString stringWithUTF8String:PyString_AS_STRING(object)];
    } else if (PyUnicode_Check(object)) {
        PyObject *str = PyUnicode_AsUTF8String(object);
        NSString *ret = PyObjectToObjc(str);
        Py_DECREF(str);
        return ret;
    } else if (PyInt_Check(object)) {
        return [NSNumber numberWithLong:PyInt_AS_LONG(object)];
    } else if (PyLong_Check(object)) {
        return [NSNumber numberWithLong:PyLong_AsLong(object)];
    } else if (PyList_Check(object) || PyTuple_Check(object)) {
        int isTuple = PyTuple_Check(object);
        Py_ssize_t len = isTuple ? PyTuple_Size(object) : PyList_Size(object);
        NSMutableArray *list = [[NSMutableArray alloc] initWithCapacity:len];
        Py_ssize_t i = 0;

        // Create function pointer to the correct GetItem function
        PyObject *(*GetArrayItem)(PyObject *, Py_ssize_t);
        GetArrayItem = isTuple ? &PyTuple_GetItem : &PyList_GetItem;

        for (i = 0; i < len; i++) {
            id objcObject = PyObjectToObjc(GetArrayItem(object, i));
            if (objcObject != nil) {
                [list addObject:objcObject];
            } else {
                NSLog(@"Error in PyObjectToObjc: Item %i was nil in list %@",
                      i, list);
            }
        }

        // Create non-mutable copy of array if list is tuple
        if (isTuple) {
            NSArray *copy = [list copy];
            [list release];
            return [copy autorelease];
        }

        return [list autorelease];
    } else if (PyAnySet_Check(object)) {
        Py_ssize_t len = PySet_Size(object);
        NSMutableArray *set = [[NSMutableArray alloc] initWithCapacity:len];
        unsigned i = 0;

        for (i = 0; i < len; i++) {
            PyObject *poppedObj = PySet_Pop(object);
            id objcObject = PyObjectToObjc(poppedObj);
            Py_DECREF(poppedObj);
            if (objcObject != nil) {
                [set addObject:objcObject];
            } else {
                NSLog(@"Error in PyObjectToObjc: Item %i was nil in set %@",
                      i, set);
            }
        }

        NSArray *copy = [set copy];
        [set release];
        return [copy autorelease];
    } else if (PyDict_Check(object)) {
        PyObject *pyKeys = PyDict_Keys(object);
        id keys = PyObjectToObjc(pyKeys);
        Py_DECREF(pyKeys);

        PyObject *pyValues = PyDict_Values(object);
        id values = PyObjectToObjc(pyValues);
        Py_DECREF(pyValues);

        return [NSDictionary dictionaryWithObjects:values forKeys:keys];
    } else if (PyBool_Check(object)) {
        return [NSNumber numberWithBool:object == Py_True];
    } else if (PyFloat_Check(object)) {
        return [NSNumber numberWithDouble:PyFloat_AsDouble(object)];
    } else {
        NSLog(@"Could not convert PyObject; returning nil.");
        return nil;
    }
}
