import sys
# vim set fileencoding=utf-8
"""
A regex parser that works by generating a dictionary of a
regex with build_regex and then parsing it with parse_word.

Currently only accepts formal regex notation.

Copyright © 2009 Michael Sanders. Released under the MIT License.
"""

def to_unicode(obj, encoding='utf-8'):
    """Converts a string to unicode if it's not already."""
    if isinstance(obj, basestring) and not isinstance(obj, unicode):
        obj = unicode(obj, encoding)
    return obj

def parse_word(transitions, word):
    """
    Returns a tuple of whether the word was accepted & the ending states the
    ε-NFA was in.
    """
    word = to_unicode(word)
    current_states = [0]
    accepting_state = len(transitions) - 1

    for char in word:
        new_states = set([])
        for state in current_states:
            new_states.update(get_states(transitions, char, current_states))
        if not new_states:
            return (-2, current_states, char)
        current_states = list(new_states)

    # Once at the end-of-string, check for possible epsilon states
    for state in current_states:
        for epsilon_state in list(get_states(transitions, u'ε', [state])):
            if not epsilon_state in current_states:
                current_states.append(epsilon_state)

    return (0 if accepting_state in current_states else -1, current_states, '')

def get_states(transitions, char, start_states):
    """
    Returns a set of states arrived at after transitioning from the
    supplied start states via the supplied character.
    (Epsilon states are appropriately recursively traversed.)
    """
    end_states = set([])

    for state in start_states:
        if transitions[state].has_key(u'ε'):
            end_states.update(get_states(transitions, char,
                                         transitions[state][u'ε']))
        if transitions[state].has_key(char):
            end_states.update(set(transitions[state][char]))

    return end_states

def shift_states(num, dfa):
    """Shifts all destinations for transitions in a DFA by the given number."""
    # Aah! SO inefficient!
    for transitions in dfa:
        for char in transitions:
            for i in range(len(transitions[char])):
                transitions[char][i] += num
    return dfa

def concat_NFAs(nfa1, nfa2):
    """Concatenates nfa1 with nfa2 and returns the result."""
    j = len(nfa1) - 1
    del nfa1[j] # It is assumed that each of our NFA's have an ending accepting
                # state with no transitions, so this is used instead of an ugly
                # ε-transition.
    shift_states(j, nfa2)
    nfa1.extend(nfa2)
    return nfa1

def insert_state(new_states, pos, nfa):
    """
    Inserts given states into the given position onto an NFA, updating the
    numbers in the transitions accordingly.
    """
    shift_states(1, nfa)
    nfa.insert(pos, new_states)
    return nfa

def union_NFAs(nfa1, nfa2):
    """Unions nfa1 with nfa2 and returns the result."""
    nfa1_len = len(nfa1) + 1
    nfa2_len = len(nfa2)
    new_nfa = insert_state({u'ε': [1, nfa1_len]}, 0, nfa1)

    shift_states(nfa1_len, nfa2)

    nfa2[-1][u'ε'] = [nfa1_len + nfa2_len]
    new_nfa[-1][u'ε'] = [nfa1_len + nfa2_len]
    new_nfa.extend(nfa2)
    new_nfa.append({})
    return new_nfa

def build_regex(infix):
    """Infix wrapper for build_postfix_regex()."""
    postfix = infix_to_postfix(infix)
    print postfix.encode('utf-8')
    return build_postfix_regex(postfix) if postfix else []

def build_postfix_regex(postfix):
    """
    Converts a formal postfix regular expression to an ε-NFA returned as
    a dictionary.
    """
    nfa_stack = []
    postfix = to_unicode(postfix)

    for char in postfix:
        if char == '&': # Concatenation
            last_nfa = nfa_stack.pop()
            nfa_stack[-1] = concat_NFAs(nfa_stack[-1], last_nfa)
        elif char == '*': # Kleene Closure
            last_nfa = nfa_stack[-1]
            nfa_len = len(last_nfa)
            last_nfa[-1][u'ε'] = [0, nfa_len]
            last_nfa.append({})
            insert_state({u'ε': [1, nfa_len + 1]}, 0, last_nfa)
        elif char == '+': # Union
            last_nfa = nfa_stack.pop()
            nfa_stack[-1] = union_NFAs(nfa_stack[-1], last_nfa)
        else:
            nfa_stack.append([{char: [1]}, {}])

    # str() doesn't seem to like unicode for some reason
    print str(nfa_stack).replace(r'\u03b5', 'ε')
    return nfa_stack[0] if nfa_stack else []

def concat(opstack, output):
    """
    Adds concatenation operator to the stack for infix to postfix conversion.
    """
    while opstack and opstack[-1] == '*':
        output += opstack.pop()
    opstack.append('&')

def implicit_concat(opstack, output, nextchar):
    """Adds concatenation operator to the stack only when necessary."""
    if nextchar and nextchar != '+' and nextchar != '&' and nextchar != '*' \
                and nextchar != ')':
        concat(opstack, output)

def infix_to_postfix(infix):
    """
    Converts an infix string to a postfix string; returns empty string on error
    """
    infix = to_unicode(infix)
    opstack = []
    output = ''
    length = len(infix)
    i = 0

    for char in infix:
        nextchar = infix[i + 1] if i + 1 < length else ''
        if char == '&': # Concatenation
            concat(opstack, output)
        elif char == '*': # Kleene Closure
            implicit_concat(opstack, output, nextchar)
            output += '*'
        elif char == '+': # Union
            while opstack and (opstack[-1] == '*' or opstack[-1] == '&'):
                output += opstack.pop()
            opstack.append('+')
        elif char == '(':
            opstack.append('(')
        elif char == ')':
            found_match = False
            while opstack:
                item = opstack.pop()
                if item == '(':
                    found_match = True
                    break
                else:
                    output += item
            if not found_match:
                print 'Mismatched parentheses'
                return ''
            else:
                implicit_concat(opstack, output, nextchar)
        else:
            implicit_concat(opstack, output, nextchar)
            output += char
        i += 1

    if not output:
        return '' # Return error if no non-operator characters are found

    for operator in opstack:
        if operator == '(' or operator == ')':
            print 'Mismatched parentheses'
            return ''
        output += operator

    return output

def main(regex=None, word=None):
    if regex is None: regex = "(01|10)(01|10)*" # 01.10.+01.10.+*.
    if word is None: word = "0110"
    print 'regex: ' + regex

    transitions = build_regex(regex)
    parsed_word, end_states, end_char = parse_word(transitions, word)

    if word == "": word = "ε"
    if parsed_word == 0:
        print word + " was accepted! (states " + str(end_states) + ")"
    elif parsed_word == -1:
        print word + " was NOT accepted. (states " + str(end_states) + ")"
    else:
        print word + " died on states " + str(end_states) + " on char " + end_char

if __name__ == "__main__":
    arglen = len(sys.argv)
    main(sys.argv[2] if arglen > 2 else None,
         sys.argv[1] if arglen > 1 else None)
